@@ -30,6 +30,15 @@
 #define N_ELEMENTS(array) (sizeof(array) / sizeof(array[0]))
 #define HANGUL_KEYBOARD_TABLE_SIZE 0x80
 
+typedef void   (*HangulOnTranslate)  (HangulInputContext*,
+				      int,
+				      ucschar*,
+				      void*);
+typedef bool   (*HangulOnTransition) (HangulInputContext*,
+				      ucschar,
+				      const ucschar*,
+				      void*);
+
 typedef struct _HangulCombinationItem HangulCombinationItem;
 
 struct _HangulKeyboard {
@@ -63,13 +72,20 @@ struct _HangulInputContext {
     const HangulCombination* combination;
 
     HangulBuffer buffer;
-    HangulICFilter filter;
-    void *filter_data;
     int output_mode;
 
     ucschar preedit_string[64];
     ucschar commit_string[64];
     ucschar flushed_string[64];
+
+    HangulOnTranslate   on_translate;
+    void*               on_translate_data;
+
+    HangulOnTransition  on_transition;
+    void*               on_transition_data;
+
+    HangulICFilter filter;
+    void *filter_data;
 };
 
 #include "hangulkeyboard.h"
@@ -274,6 +290,24 @@ hangul_buffer_is_empty(HangulBuffer *buffer)
 	   buffer->jongseong == 0;
 }
 
+static bool
+hangul_buffer_has_choseong(HangulBuffer *buffer)
+{
+    return buffer->choseong != 0;
+}
+
+static bool
+hangul_buffer_has_jungseong(HangulBuffer *buffer)
+{
+    return buffer->jungseong != 0;
+}
+
+static bool
+hangul_buffer_has_jongseong(HangulBuffer *buffer)
+{
+    return buffer->jongseong != 0;
+}
+
 static void
 hangul_buffer_push(HangulBuffer *buffer, ucschar ch)
 {
@@ -441,41 +475,41 @@ hangul_buffer_backspace(HangulBuffer *buffer)
 }
 
 static inline bool
-hangul_ic_push(HangulInputContext *hic, ucschar ch)
+hangul_ic_push(HangulInputContext *hic, ucschar c)
 {
     ucschar buf[64] = { 0, };
-    if (hic->filter != NULL) {
+    if (hic->on_transition != NULL) {
 	ucschar cho, jung, jong;
-	if (hangul_is_choseong(ch)) {
-	    cho  = ch;
+	if (hangul_is_choseong(c)) {
+	    cho  = c;
 	    jung = hic->buffer.jungseong;
 	    jong = hic->buffer.jongseong;
-	} else if (hangul_is_jungseong(ch)) {
+	} else if (hangul_is_jungseong(c)) {
 	    cho  = hic->buffer.choseong;
-	    jung = ch;
+	    jung = c;
 	    jong = hic->buffer.jongseong;
-	} else if (hangul_is_jongseong(ch)) {
+	} else if (hangul_is_jongseong(c)) {
 	    cho  = hic->buffer.choseong;
 	    jung = hic->buffer.jungseong;
-	    jong = ch;
+	    jong = c;
 	} else {
 	    hangul_ic_flush_internal(hic);
 	    return false;
 	}
 
 	hangul_jaso_to_string(cho, jung, jong, buf, N_ELEMENTS(buf));
-	if (!hic->filter(buf, cho, jung, jong, hic->filter_data)) {
+	if (!hic->on_transition(hic, c, buf, hic->on_transition_data)) {
 	    hangul_ic_flush_internal(hic);
 	    return false;
 	}
     } else {
-	if (!hangul_is_jaso(ch)) {
+	if (!hangul_is_jaso(c)) {
 	    hangul_ic_flush_internal(hic);
 	    return false;
 	}
     }
 
-    hangul_buffer_push(&hic->buffer, ch);
+    hangul_buffer_push(&hic->buffer, c);
     return true;
 }
 
@@ -767,20 +801,22 @@ hangul_ic_process_jaso(HangulInputContext *hic, ucschar ch)
 bool
 hangul_ic_process(HangulInputContext *hic, int ascii)
 {
-    ucschar ch;
+    ucschar c;
 
     if (hic == NULL)
 	return false;
 
-    ch = hangul_keyboard_get_value(hic->keyboard, ascii);
-
     hic->preedit_string[0] = 0;
     hic->commit_string[0] = 0;
 
+    c = hangul_keyboard_get_value(hic->keyboard, ascii);
+    if (hic->on_translate != NULL)
+	hic->on_translate(hic, ascii, &c, hic->on_translate_data);
+
     if (hangul_keyboard_get_type(hic->keyboard) == HANGUL_KEYBOARD_TYPE_JAMO)
-	return hangul_ic_process_jamo(hic, ch);
+	return hangul_ic_process_jamo(hic, c);
     else
-	return hangul_ic_process_jaso(hic, ch);
+	return hangul_ic_process_jaso(hic, c);
 }
 
 const ucschar*
@@ -969,6 +1005,24 @@ hangul_ic_is_empty(HangulInputContext *hic)
     return hangul_buffer_is_empty(&hic->buffer);
 }
 
+bool
+hangul_ic_has_choseong(HangulInputContext *hic)
+{
+    return hangul_buffer_has_choseong(&hic->buffer);
+}
+
+bool
+hangul_ic_has_jungseong(HangulInputContext *hic)
+{
+    return hangul_buffer_has_jungseong(&hic->buffer);
+}
+
+bool
+hangul_ic_has_jongseong(HangulInputContext *hic)
+{
+    return hangul_buffer_has_jongseong(&hic->buffer);
+}
+
 void
 hangul_ic_set_output_mode(HangulInputContext *hic, int mode)
 {
@@ -978,14 +1032,47 @@ hangul_ic_set_output_mode(HangulInputContext *hic, int mode)
     hic->output_mode = mode;
 }
 
+void
+hangul_ic_connect_translate (HangulInputContext* hic,
+                             HangulOnTranslate callback,
+                             void* user_data)
+{
+    if (hic != NULL) {
+	hic->on_translate      = callback;
+	hic->on_translate_data = user_data;
+    }
+}
+
+void
+hangul_ic_connect_transition(HangulInputContext* hic,
+                             HangulOnTransition callback,
+                             void* user_data)
+{
+    if (hic != NULL) {
+	hic->on_transition      = callback;
+	hic->on_transition_data = user_data;
+    }
+}
+
+void hangul_ic_connect_callback(HangulInputContext* hic, const char* event,
+				void* callback, void* user_data)
+{
+    if (hic == NULL || event == NULL)
+	return;
+
+    if (strcasecmp(event, "translate") == 0) {
+	hic->on_translate      = (HangulOnTranslate)callback;
+	hic->on_translate_data = user_data;
+    } else if (strcasecmp(event, "transition") == 0) {
+	hic->on_transition      = (HangulOnTransition)callback;
+	hic->on_transition_data = user_data;
+    }
+}
+
 void hangul_ic_set_filter(HangulInputContext *hic,
 			  HangulICFilter func, void *user_data)
 {
-    if (hic == NULL)
-	return;
-
-    hic->filter = func;
-    hic->filter_data = user_data;
+    return;
 }
 
 void
@@ -1048,7 +1135,6 @@ hangul_ic_new(const char* keyboard)
 
     hangul_ic_set_output_mode(hic, HANGUL_OUTPUT_SYLLABLE);
     hangul_ic_select_keyboard(hic, keyboard);
-    hangul_ic_set_filter(hic, NULL, NULL);
 
     hangul_buffer_clear(&hic->buffer);
 
@@ -1056,6 +1142,12 @@ hangul_ic_new(const char* keyboard)
     hic->commit_string[0] = 0;
     hic->flushed_string[0] = 0;
 
+    hic->on_translate      = NULL;
+    hic->on_translate_data = NULL;
+
+    hic->on_transition      = NULL;
+    hic->on_transition_data = NULL;
+
     return hic;
 }
 
